/**
 *  Pcap capture multiplexor
 *  
 *  @author     ETSI / STF424
 *  @version    $URL: $
 *              $Id: PcapMultiplexer.java,v 1.1 2014/11/20 13:40:57 bgr Exp $
 */
package org.etsi.its.adapter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jpcap.JpcapCaptor;
import jpcap.JpcapSender;
import jpcap.NetworkInterface;
import jpcap.PacketReceiver;
import jpcap.packet.EthernetPacket;
import jpcap.packet.Packet;

import org.etsi.adapter.TERFactory;
import org.etsi.common.ByteHelper;
import org.etsi.its.adapter.layers.EthernetLayer;
import org.etsi.its.adapter.layers.Layer;
import org.etsi.ttcn.tci.CharstringValue;

public class PcapMultiplexer implements Runnable {

	/**
	 * Unique instance of the factory
	 */
	private static final PcapMultiplexer instance = new PcapMultiplexer();

	private static byte[] MAC_BROADCAST = new byte[]{(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF};

	private StringBuilder errbuf = new StringBuilder();     // For any error msgs  

	private boolean runThread;

	private PcapMultiplexer() {

		filter = "";

		// Obtain the list of network interfaces
		List<NetworkInterface> alldevs = new ArrayList<NetworkInterface>(); // Will be filled with NICs  


		Collections.addAll(alldevs, JpcapCaptor.getDeviceList());

		if (alldevs.isEmpty()) {  
			System.err.printf("Can't read list of devices, error is %s", errbuf.toString());  
			return;  
		}  

		// Find the right interface        
		int ifaceIndex = 0;
		String expectedIface = ((CharstringValue)TERFactory.getInstance().getTaParameter("LocalEthernetMAC")).getString().toLowerCase(); 
		for( ; ifaceIndex < alldevs.size(); ifaceIndex++) {
			if (expectedIface.equalsIgnoreCase(ByteHelper.byteArrayToString(alldevs.get(ifaceIndex).mac_address))) {
				// Interface found
				break;
			}
		}
		// Check result
		if (ifaceIndex == alldevs.size()) {
			throw new RuntimeException(String.format("EthernetLayer.register: Network interface %s not found", expectedIface));
		}

		device = alldevs.get(ifaceIndex);
		try {
			jpcapSender = JpcapSender.openDevice(device);
		} catch (IOException e) {
			throw new RuntimeException(String.format("JpcapSender.openDevice: Network interface %s can not open for sending", expectedIface));
		}
	}

	/**
	 * Gets the unique factory instance
	 * @return PcapMultiplexer instance
	 */
	public static PcapMultiplexer getInstance(){
		return instance;
	}

	public synchronized void register(Layer client, byte[] macAddress, short frameType) {

		if(clientsToMacs.isEmpty()) {

			// Open interface 

			int snaplen = 64 * 1024;            // Capture all packets, no truncation  
			int timeout = 10;                   // 10 millis only working for linux
			try {
				pcap = JpcapCaptor.openDevice(device, snaplen, true, timeout);
			} catch (IOException e) {
				throw new RuntimeException(String.format("JpcapCaptor.openDevice: Network interface %s can not open for sending", device.name));
			}  

			if (pcap == null) {  
				System.err.printf("Error while opening device for capture: "  
						+ errbuf.toString());  
				return;  
			}  
			captureThread = new Thread(this);
			captureThread.start();   
			filter = "";
		}
		else {
			filter = filter + " and ";
		}

		// Update Filter
		String strMacAddress = String.format("%02x", macAddress[0]);
		for(int i=1; i < macAddress.length; i++) {
			strMacAddress += String.format(":%02x", macAddress[i]);
		}

		filter = filter + "not ether src " + strMacAddress;

		try {
			pcap.setFilter(filter,true);
		} catch (IOException e) {
			throw new RuntimeException(String.format("JpcapCaptor.setFilter: Filter is not applicable: %s", filter));
		}

		// Register client
		clientsToMacs.put(client.toString(), macAddress);
		clientsToLayers.put(client.toString(), client);
		clientsToFrameTypes.put(client.toString(), frameType);
	}

	public synchronized void unregister(Layer client) {
		if(clientsToMacs.containsKey(client.toString())) {
			clientsToMacs.remove(client.toString());
			clientsToFrameTypes.remove(client.toString());
			clientsToLayers.remove(client.toString());

			if(clientsToMacs.isEmpty()) {
				pcap.breakLoop();
				try {
					captureThread.join();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				pcap.close();
			}
		}
	}

	/**
	 * Thread function for jpcap capture loop
	 * @see java.lang.Runnable#run()
	 */
	@Override
	public void run() {

		runThread = true;

		pcap.loopPacket(-1, new PacketReceiver() {
			@Override
			public void receivePacket(Packet packet) {


				if (packet == Packet.EOF) {
					runThread = false;
					pcap.breakLoop();
				}



				if (isThreadRunning()) {
					EthernetPacket ethPacket = null;

					if(packet.datalink instanceof EthernetPacket){

						ethPacket = (EthernetPacket)packet.datalink;

						Map<String, Object> lowerInfo = new HashMap<String, Object>();

						byte[] dst = ethPacket.dst_mac;
						short frameType = ethPacket.frametype;
						byte[] data = packet.data;


						lowerInfo.put(EthernetLayer.LINK_LAYER_DESTINATION, dst);

						// Dispatch
						for (String mapKey : clientsToMacs.keySet()) {
							if(frameType == clientsToFrameTypes.get(mapKey)) {
								if(Arrays.equals(dst, MAC_BROADCAST)
										|| Arrays.equals(dst, clientsToMacs.get(mapKey))) {

									lowerInfo.put(Layer.RECEPTION_TIMESTAMP, packet.sec*1000000 + packet.usec);
									clientsToLayers.get(mapKey).receive(data, lowerInfo);
								}
							}
						}

					}

				}

			}
		});
	}

	public byte[] sendPacket(Layer client, byte[] dest, byte[] payload) {

		if(clientsToMacs.containsKey(client.toString())) {


			Packet rawP = new Packet();
			EthernetPacket ethP = new EthernetPacket();

			ethP.dst_mac = dest;
			ethP.src_mac = clientsToMacs.get(client.toString());

			ethP.frametype = (short) clientsToFrameTypes.get(client.toString());

			rawP.datalink = ethP;

			rawP.data = payload.clone();

			jpcapSender.sendPacket(rawP);

			byte[] packet = ByteHelper.concat(rawP.header, rawP.data);

			return packet;
		}
		return null;
	}

	private synchronized boolean isThreadRunning(){
		return runThread;
	}

	public synchronized void stopThread(){
		runThread = false;
	}

	/**
	 * Jpcap capture device
	 */
	private JpcapCaptor pcap;

	/**
	 * Jpcap capture thread instance.
	 */
	private Thread captureThread;

	NetworkInterface device;
	JpcapSender jpcapSender;
	private String filter;
	private Map<String, byte[]> clientsToMacs = new HashMap<String, byte[]>();
	private Map<String, Short> clientsToFrameTypes = new HashMap<String, Short>();
	private HashMap<String, Layer> clientsToLayers = new HashMap<String, Layer>();
}   